# Copyright (C) 2014-2019  Sutou Kouhei <kou@clear-code.com>
# Copyright (C) 2019  Horimoto Yasuhiro <horimoto@clear-code.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

require "English"
require "find"
require "pp"
require "optparse"
require "json"

require "diff/lcs"
require "diff/lcs/hunk"
require "groonga/command/parser"

require "groonga-query-log/formattable"
require "groonga-query-log/version"

module GroongaQueryLog
  module Command
    class FormatRegressionTestLogs
      include Formattable

      def initialize(options={})
        @output = options[:output] || $stdout
      end

      def run(command_line)
        parser = OptionParser.new
        parser.banner += " PATH1 PATH2 ..."
        parser.version = VERSION
        paths = parser.parse!(command_line)

        if paths.empty?
          format_log($stdin, "-")
        else
          paths.each do |path|
            if File.directory?(path)
              Find.find(path) do |sub_path|
                next unless File.file?(sub_path)
                File.open(sub_path, encoding: "UTF-8") do |file|
                  format_log(file, sub_path)
                end
              end
            else
              File.open(path, encoding: "UTF-8") do |file|
                format_log(file, path)
              end
            end
          end
        end
        true
      end

      private
      def format_log(input, path)
        command = nil
        response_old = nil
        response_new = nil
        backtrace = []
        error_message = nil
        elapsed_time_old = nil
        elapsed_time_new = nil
        elapsed_time_ratio = nil
        elapsed_times_old = nil
        elapsed_times_new = nil

        input.each_line do |line|
          unless line.valid_encoding?
            @output.puts("invalid encoding line")
            @output.puts("#{path}:#{input.lineno}:#{line}")
            next
          end
          case line
          when /\Acommand: /
            command = $POSTMATCH.chomp
          when /\Aresponse1: /
            response_old = $POSTMATCH.chomp
          when /\Aresponse2: /
            response_new = $POSTMATCH.chomp
            next unless valid_entry?(command, response_old, response_new)
            report_diff(command, response_old, response_new)
          when /\Aerror: /
            error_message = $POSTMATCH.chomp
            report_error(command, error_message, backtrace)
            backtrace.clear
          when /\Abacktrace: /
            backtrace.unshift($POSTMATCH.chomp)
          when /\Aelapsed_time_old: /
            elapsed_time_old = Float($POSTMATCH.chomp)
          when /\Aelapsed_time_new: /
            elapsed_time_new = Float($POSTMATCH.chomp)
          when /\Aelapsed_times_old: /
            elapsed_times_old = $POSTMATCH.chomp.split.collect do |value|
              Float(value)
            end
          when /\Aelapsed_times_new: /
            elapsed_times_new = $POSTMATCH.chomp.split.collect do |value|
              Float(value)
            end
          when /\Aelapsed_time_ratio: /
            elapsed_time_ratio = Float($POSTMATCH.chomp)
            report_slow(command,
                        elapsed_time_old,
                        elapsed_time_new,
                        elapsed_times_old,
                        elapsed_times_new,
                        elapsed_time_ratio)
          end
        end
      end

      def valid_entry?(command, response_old, response_new)
        valid = true
        unless validate_response(command, "old", response_old)
          valid = false
        end
        unless validate_response(command, "new", response_new)
          valid = false
        end
        valid
      end

      def validate_response(command, type, response)
        if response.nil?
          @output.puts(command)
          @output.puts("no #{type} response")
          return false
        end

        begin
          JSON.parse(response)
        rescue JSON::ParserError
          @output.puts(command)
          @output.puts("failed to parse #{type} response: #{$!.message}")
          @output.puts(response)
          return false
        end

        true
      end

      def report_diff(command, response_old, response_new)
        return if response_old == response_new

        report_command(command)

        lines_old = response_to_lines(response_old)
        lines_new = response_to_lines(response_new)
        diffs = Diff::LCS.diff(lines_old, lines_new)

        @output.puts("--- old")
        @output.puts("+++ new")

        old_hunk = nil
        n_lines = 3
        format = :unified
        file_length_difference = 0
        diffs.each do |piece|
          begin
            hunk = Diff::LCS::Hunk.new(lines_old,
                                       lines_new,
                                       piece,
                                       n_lines,
                                       file_length_difference)
            file_length_difference = hunk.file_length_difference

            next unless old_hunk

            if (n_lines > 0) && hunk.overlaps?(old_hunk)
              hunk.merge(old_hunk)
            else
              @output.puts(old_hunk.diff(format))
            end
          ensure
            old_hunk = hunk
          end
        end

        if old_hunk
          @output.puts(old_hunk.diff(format))
        end
      end

      def response_to_lines(response)
        PP.pp(JSON.parse(response), "").lines.collect(&:chomp)
      end

      def report_error(command, message, backtrace)
        report_command(command)
        @output.puts("Error: #{message}")
        @output.puts("Backtrace:")
        @output.puts(backtrace)
      end

      def report_slow(command,
                      elapsed_time_old,
                      elapsed_time_new,
                      elapsed_times_old,
                      elapsed_times_new,
                      elapsed_time_ratio)
        report_command(command)
        elapsed_times_old ||= [elapsed_time_old]
        elapsed_times_new ||= [elapsed_time_new]
        @output.puts("Slow:")
        @output.puts("  Old: %s (%s)" % [
                       format_elapsed_time(elapsed_time_old),
                       format_elapsed_times(elapsed_times_old),
                     ])
        @output.puts("  New: %s (%s)" % [
                       format_elapsed_time(elapsed_time_new),
                       format_elapsed_times(elapsed_times_new),
                     ])
        @output.puts("  Ratio: +%.1f%%" % ((elapsed_time_ratio * 100) - 100))
      end

      def report_command(command)
        @output.puts("Command:")
        @output.puts(command)
        parsed_command = Groonga::Command::Parser.parse(command)
        @output.puts("Name: #{parsed_command.name}")
        @output.puts("Arguments:")
        sorted_arguments = parsed_command.arguments.sort_by do |key, value|
          key
        end
        sorted_arguments.each do |key, value|
          @output.puts("  #{key}: #{value}")
        end
      end
    end
  end
end
